"""Mixin class to handle random number generation."""

from __future__ import annotations

from jax import Array

from gwmock_pop.rng.rng import RNGManager


class RandomMixin:
    """Mixin class to handle random number generation."""

    def __init__(self, *args, seed: int | None = None, **kwargs) -> None:
        """Initialize the instance.

        Args:
            *args: Positional arguments.
            seed: Seed to initialize a random number generator.
            **kwargs: Keyword arguments.
        """
        self._rng_manager = RNGManager(seed=seed)
        super().__init__(*args, **kwargs)

    @property
    def rng_manager(self) -> RNGManager:
        """Get the RNG manager.

        Returns:
            RNG manager.
        """
        return self._rng_manager

    @property
    def rng_key_data(self) -> Array:
        """Get the key data of the random number generator.

        Returns:
            Key data of the random number generator.
        """
        return self._rng_manager.key_data



"""Unit tests for gwmock_pop.constants."""

from gwmock_pop.constants import SPEED_OF_LIGHT

SPEED_OF_LIGHT_REFERENCE = 299792458.0


class TestConstants:
    """Test the constants."""

    def test_speed_of_light(self):
        """Compare the speed of light with the expected value."""
        assert SPEED_OF_LIGHT == SPEED_OF_LIGHT_REFERENCE



"""Tests for graph sampler helper functions."""

from __future__ import annotations

from gwmock_pop.graph.sampler import extract_sampler_dependencies


class TestExtractSamplerDependencies:
    """Test suite for extract_sampler_dependencies function."""

    def test_empty_dict_returns_empty_set(self) -> None:
        """Test that empty dict returns empty set."""
        result = extract_sampler_dependencies({})
        assert result == set()

    def test_function_sampler_with_list_arguments(self) -> None:
        """Test extracting dependencies from function sampler with list arguments."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": ["@mass_1", "@mass_2"],
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"mass_1", "mass_2"}

    def test_function_sampler_with_dict_arguments(self) -> None:
        """Test extracting dependencies from function sampler with dict arguments."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": {
                "param1": "@mass_1",
                "param2": "@mass_2",
            },
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"mass_1", "mass_2"}

    def test_function_sampler_with_depends_on(self) -> None:
        """Test extracting dependencies using depends_on key."""
        sampler_spec = {
            "function": "some_sampler",
            "depends_on": ["@param1", "@param2"],
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"param1", "param2"}

    def test_function_sampler_prefers_arguments_over_depends_on(self) -> None:
        """Test that arguments key takes precedence over depends_on."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": ["@arg1"],
            "depends_on": ["@dep1"],
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"arg1"}

    def test_non_function_sampler_uses_extract_dependencies_from_spec(self) -> None:
        """Test that non-function sampler uses extract_dependencies_from_spec."""
        sampler_spec = {
            "min": 1.0,
            "max": 100.0,
            "expression": "@base_value",
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"base_value"}

    def test_simple_dict_no_function(self) -> None:
        """Test simple dict without function key."""
        sampler_spec = {"min": 1.0, "max": 100.0}
        result = extract_sampler_dependencies(sampler_spec)
        assert result == set()

    def test_nested_dict_in_sampler(self) -> None:
        """Test nested dict in sampler spec."""
        sampler_spec = {
            "nested": {
                "expression": "@nested_value",
            },
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"nested_value"}

    def test_list_arguments_with_empty_strings(self) -> None:
        """Test list arguments with empty strings."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": ["@valid", "", "@"],
        }
        result = extract_sampler_dependencies(sampler_spec)
        # Empty strings and '@' without name are filtered out
        assert result == {"valid"}

    def test_dict_arguments_with_empty_values(self) -> None:
        """Test dict arguments with empty values."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": {
                "valid": "@param",
                "empty": "",
                "not_starting_with_at": "value",
            },
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"param"}

    def test_arguments_as_string_raises_error(self) -> None:
        """Test that string arguments raises error (not expected behavior)."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": "@invalid",
        }
        # This should handle gracefully - currently returns empty set
        result = extract_sampler_dependencies(sampler_spec)
        assert result == set()

    def test_depends_on_as_string_raises_error(self) -> None:
        """Test that string depends_on raises error (not expected behavior)."""
        sampler_spec = {
            "function": "some_sampler",
            "depends_on": "@invalid",
        }
        # This should handle gracefully - currently returns empty set
        result = extract_sampler_dependencies(sampler_spec)
        assert result == set()

    def test_complex_sampler_spec(self) -> None:
        """Test complex sampler spec with multiple features."""
        sampler_spec = {
            "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
            "arguments": {
                "alpha_1": "@alpha1_value",
                "alpha_2": 4.51,
                "transition": "@transition_mass",
                "minimum": 5.06,
                "maximum": 300.0,
                "mean_1": "@peak1_mean",
                "sigma_1": "@peak1_sigma",
                "mean_2": "@peak2_mean",
                "sigma_2": "@peak2_sigma",
                "taper_range": "@taper_value",
                "lambda_0": "@lambda0",
                "lambda_1": "@lambda1",
            },
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {
            "alpha1_value",
            "transition_mass",
            "peak1_mean",
            "peak1_sigma",
            "peak2_mean",
            "peak2_sigma",
            "taper_value",
            "lambda0",
            "lambda1",
        }

    def test_depends_on_list(self) -> None:
        """Test depends_on with list of references."""
        sampler_spec = {
            "function": "some_sampler",
            "depends_on": ["@dep1", "@dep2", "@dep3"],
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"dep1", "dep2", "dep3"}

    def test_mixed_arguments_and_depends_on(self) -> None:
        """Test when both arguments and depends_on are present."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": ["@arg1", "@arg2"],
            "depends_on": ["@dep1", "@dep2"],
        }
        # Arguments should take precedence
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"arg1", "arg2"}

    def test_empty_arguments_list(self) -> None:
        """Test empty arguments list."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": [],
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == set()

    def test_empty_depends_on_list(self) -> None:
        """Test empty depends_on list."""
        sampler_spec = {
            "function": "some_sampler",
            "depends_on": [],
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == set()

    def test_none_arguments(self) -> None:
        """Test None arguments."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": None,
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == set()

    def test_none_depends_on(self) -> None:
        """Test None depends_on."""
        sampler_spec = {
            "function": "some_sampler",
            "depends_on": None,
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == set()

    def test_depends_on_used_when_arguments_missing(self) -> None:
        """Test that depends_on is used when arguments is missing."""
        sampler_spec = {
            "function": "some_sampler",
            "depends_on": ["@dep1", "@dep2"],
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"dep1", "dep2"}

    def test_function_sampler_with_mixed_arguments(self) -> None:
        """Test function sampler with mixed string and non-string arguments."""
        sampler_spec = {
            "function": "some_sampler",
            "arguments": ["@param1", 42, "@param2", None],
        }
        result = extract_sampler_dependencies(sampler_spec)
        assert result == {"param1", "param2"}

    def test_nested_function_sampler(self) -> None:
        """Test nested function sampler."""
        sampler_spec = {
            "function": "outer_sampler",
            "arguments": {
                "inner": {
                    "function": "inner_sampler",
                    "arguments": ["@inner_dep"],
                },
                "outer_dep": "@outer_dep",
            },
        }
        result = extract_sampler_dependencies(sampler_spec)
        # Should extract from the top level only
        assert result == {"outer_dep"}



"""Tests for graph build functions."""

from __future__ import annotations

import networkx as nx
import pytest
from networkx import DiGraph

from gwmock_pop.graph.build import add_dependencies_to_graph, build_dependency_graph


class TestAddDependenciesToGraph:
    """Test suite for add_dependencies_to_graph function."""

    def test_empty_dependencies_does_nothing(self) -> None:
        """Test that empty dependencies don't add edges."""
        graph = DiGraph()
        graph.add_node("param1")
        add_dependencies_to_graph(graph=graph, dependencies=set(), parameter_name="param1")
        assert len(graph.edges()) == 0

    def test_single_dependency_creates_edge(self) -> None:
        """Test that single dependency creates one edge."""
        graph = DiGraph()
        graph.add_node("param1")
        graph.add_node("dependency")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dependency"},
            parameter_name="param1",
        )
        assert len(graph.edges()) == 1
        assert ("dependency", "param1") in graph.edges()

    def test_multiple_dependencies_create_multiple_edges(self) -> None:
        """Test that multiple dependencies create multiple edges."""
        graph = DiGraph()
        graph.add_node("param1")
        graph.add_node("dep1")
        graph.add_node("dep2")
        graph.add_node("dep3")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep1", "dep2", "dep3"},
            parameter_name="param1",
        )
        assert len(graph.edges()) == 3  # noqa: PLR2004
        assert ("dep1", "param1") in graph.edges()
        assert ("dep2", "param1") in graph.edges()
        assert ("dep3", "param1") in graph.edges()

    def test_dependencies_added_to_existing_edges(self) -> None:
        """Test that dependencies are added to existing edges."""
        graph = DiGraph()
        graph.add_node("param1")
        graph.add_node("dep1")
        graph.add_edge("dep1", "param1")
        graph.add_node("dep2")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep2"},
            parameter_name="param1",
        )
        assert len(graph.edges()) == 2  # noqa: PLR2004
        assert ("dep1", "param1") in graph.edges()
        assert ("dep2", "param1") in graph.edges()

    def test_parameter_not_in_graph_adds_node(self) -> None:
        """Test that parameter not in graph adds the node."""
        graph = DiGraph()
        graph.add_node("other_param")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep1"},
            parameter_name="param1",
        )
        assert "param1" in graph.nodes()
        assert ("dep1", "param1") in graph.edges()

    def test_dependency_not_in_graph_creates_node(self) -> None:
        """Test that dependency not in graph adds the node."""
        graph = DiGraph()
        graph.add_node("param1")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep1"},
            parameter_name="param1",
        )
        assert "dep1" in graph.nodes()
        assert ("dep1", "param1") in graph.edges()

    def test_self_dependency_creates_self_loop(self) -> None:
        """Test that self-dependency creates a self-loop."""
        graph = DiGraph()
        graph.add_node("param1")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"param1"},
            parameter_name="param1",
        )
        assert ("param1", "param1") in graph.edges()

    def test_duplicate_dependencies_dont_create_duplicate_edges(self) -> None:
        """Test that duplicate dependencies don't create duplicate edges."""
        graph = DiGraph()
        graph.add_node("param1")
        graph.add_node("dep1")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep1"},
            parameter_name="param1",
        )
        # NetworkX DiGraph doesn't allow duplicate edges
        assert len(graph.edges()) == 1
        assert ("dep1", "param1") in graph.edges()

    def test_order_of_dependencies_preserved(self) -> None:
        """Test that order of dependencies is preserved in iteration."""
        graph = DiGraph()
        graph.add_node("param1")
        graph.add_node("dep1")
        graph.add_node("dep2")
        graph.add_node("dep3")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep1", "dep2", "dep3"},
            parameter_name="param1",
        )
        # Check that all edges exist
        edges = list(graph.edges())
        assert len(edges) == 3  # noqa: PLR2004
        for dep in ["dep1", "dep2", "dep3"]:
            assert (dep, "param1") in edges

    def test_graph_is_directed(self) -> None:
        """Test that the graph remains directed."""
        graph = DiGraph()
        graph.add_node("param1")
        graph.add_node("dep1")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep1"},
            parameter_name="param1",
        )
        # Edge should be directed from dep1 to param1
        assert ("dep1", "param1") in graph.edges()
        # Reverse edge should not exist
        assert ("param1", "dep1") not in graph.edges()

    def test_multiple_parameters_with_dependencies(self) -> None:
        """Test adding dependencies for multiple parameters."""
        graph = DiGraph()
        graph.add_node("param1")
        graph.add_node("param2")
        graph.add_node("dep1")
        graph.add_node("dep2")
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep1"},
            parameter_name="param1",
        )
        add_dependencies_to_graph(
            graph=graph,
            dependencies={"dep2"},
            parameter_name="param2",
        )
        assert len(graph.edges()) == 2  # noqa: PLR2004
        assert ("dep1", "param1") in graph.edges()
        assert ("dep2", "param2") in graph.edges()


class TestBuildDependencyGraph:
    """Test suite for build_dependency_graph function."""

    def test_empty_config_returns_empty_graph(self) -> None:
        """Test that empty config returns empty graph."""
        graph = build_dependency_graph({})
        assert len(graph.nodes()) == 0
        assert len(graph.edges()) == 0

    def test_single_sampler_parameter(self) -> None:
        """Test building graph with single sampler parameter."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "some_sampler",
                    "arguments": [],
                },
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_1" in graph.nodes()
        assert len(graph.edges()) == 0

    def test_sampler_with_dependencies(self) -> None:
        """Test building graph with sampler that has dependencies."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "some_sampler",
                    "arguments": ["@base_mass"],
                },
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_1" in graph.nodes()
        assert "base_mass" in graph.nodes()
        assert ("base_mass", "mass_1") in graph.edges()

    def test_sampler_with_dict_arguments(self) -> None:
        """Test building graph with sampler with dict arguments."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "some_sampler",
                    "arguments": {
                        "param1": "@value1",
                        "param2": "@value2",
                    },
                },
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_1" in graph.nodes()
        assert "value1" in graph.nodes()
        assert "value2" in graph.nodes()
        assert ("value1", "mass_1") in graph.edges()
        assert ("value2", "mass_1") in graph.edges()

    def test_transform_with_string_expression(self) -> None:
        """Test building graph with transform string."""
        config = {
            "mass_ratio": {
                "transform": "@mass_2 / @mass_1",
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_ratio" in graph.nodes()
        assert "mass_1" in graph.nodes()
        assert "mass_2" in graph.nodes()
        assert ("mass_1", "mass_ratio") in graph.edges()
        assert ("mass_2", "mass_ratio") in graph.edges()

    def test_transform_with_dict(self) -> None:
        """Test building graph with transform dict."""
        config = {
            "mass_ratio": {
                "transform": {
                    "function": "some_transform",
                    "arguments": ["@input1", "@input2"],
                },
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_ratio" in graph.nodes()
        assert "input1" in graph.nodes()
        assert "input2" in graph.nodes()
        assert ("input1", "mass_ratio") in graph.edges()
        assert ("input2", "mass_ratio") in graph.edges()

    def test_condition_expression(self) -> None:
        """Test building graph with condition expression."""
        config = {
            "mass_1": {
                "condition": "@mass_1 > 5",
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_1" in graph.nodes()
        assert ("mass_1", "mass_1") in graph.edges()

    def test_condition_else_block(self) -> None:
        """Test building graph with condition_else block."""
        config = {
            "mass_1": {
                "condition_else": {
                    "value": "@fallback_value",
                },
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_1" in graph.nodes()
        assert "fallback_value" in graph.nodes()
        assert ("fallback_value", "mass_1") in graph.edges()

    def test_branches_style(self) -> None:
        """Test building graph with branches."""
        config = {
            "mass_1": {
                "branches": [
                    {
                        "condition": "@condition1",
                        "value": "@value1",
                    },
                    {
                        "condition": "@condition2",
                        "value": "@value2",
                    },
                ],
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_1" in graph.nodes()
        assert "condition1" in graph.nodes()
        assert "condition2" in graph.nodes()
        assert "value1" in graph.nodes()
        assert "value2" in graph.nodes()
        assert ("condition1", "mass_1") in graph.edges()
        assert ("condition2", "mass_1") in graph.edges()
        assert ("value1", "mass_1") in graph.edges()
        assert ("value2", "mass_1") in graph.edges()

    def test_complex_config_with_multiple_features(self) -> None:
        """Test building graph with complex config."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "some_sampler",
                    "arguments": ["@base_mass"],
                },
            },
            "mass_ratio": {
                "transform": "@mass_2 / @mass_1",
            },
            "mass_2": {
                "sampler": {
                    "function": "some_sampler",
                    "arguments": ["@base_mass"],
                },
            },
            "final_mass": {
                "transform": "@mass_1 + @mass_ratio",
            },
        }
        graph = build_dependency_graph(config)
        # Check all nodes are present
        assert "mass_1" in graph.nodes()
        assert "mass_ratio" in graph.nodes()
        assert "mass_2" in graph.nodes()
        assert "final_mass" in graph.nodes()
        assert "base_mass" in graph.nodes()

        # Check edges
        assert ("base_mass", "mass_1") in graph.edges()
        assert ("base_mass", "mass_2") in graph.edges()
        assert ("mass_1", "mass_ratio") in graph.edges()
        assert ("mass_2", "mass_ratio") in graph.edges()
        assert ("mass_1", "final_mass") in graph.edges()
        assert ("mass_ratio", "final_mass") in graph.edges()

    def test_multiple_parameters_same_dependency(self) -> None:
        """Test multiple parameters depending on same value."""
        config = {
            "mass_1": {
                "sampler": {
                    "arguments": ["@common"],
                },
            },
            "mass_2": {
                "sampler": {
                    "arguments": ["@common"],
                },
            },
        }
        graph = build_dependency_graph(config)
        assert "common" in graph.nodes()
        assert ("common", "mass_1") in graph.edges()
        assert ("common", "mass_2") in graph.edges()

    def test_circular_dependency(self) -> None:
        """Test that circular dependencies are allowed (directed graph)."""
        config = {
            "param1": {
                "transform": "@param2",
            },
            "param2": {
                "transform": "@param1",
            },
        }
        graph = build_dependency_graph(config)
        assert ("param2", "param1") in graph.edges()
        assert ("param1", "param2") in graph.edges()

    def test_sampler_and_transform_same_parameter(self) -> None:
        """Test parameter with both sampler and transform."""
        config = {
            "mass_1": {
                "sampler": {
                    "arguments": ["@base_mass"],
                },
                "transform": "@mass_1 * 2",
            },
        }
        graph = build_dependency_graph(config)
        assert "mass_1" in graph.nodes()
        assert "base_mass" in graph.nodes()
        # Sampler dependency
        assert ("base_mass", "mass_1") in graph.edges()
        # Transform dependency (self-reference)
        assert ("mass_1", "mass_1") in graph.edges()

    def test_nested_branches(self) -> None:
        """Test nested branches structure."""
        config = {
            "param1": {
                "branches": [
                    {
                        "condition": "@cond1",
                        "value": "@val1",
                        "branches": [
                            {
                                "condition": "@cond2",
                                "value": "@val2",
                            },
                        ],
                    },
                ],
            },
        }
        graph = build_dependency_graph(config)
        assert "param1" in graph.nodes()
        assert "cond1" in graph.nodes()
        assert "val1" in graph.nodes()
        assert "cond2" in graph.nodes()
        assert "val2" in graph.nodes()

    def test_graph_is_directed_acyclic_when_no_cycles(self) -> None:
        """Test that graph is directed acyclic when no cycles."""
        config = {
            "param1": {
                "sampler": {"arguments": []},
            },
            "param2": {
                "transform": "@param1",
            },
            "param3": {
                "transform": "@param2",
            },
        }
        graph = build_dependency_graph(config)
        # Check DAG structure
        assert len(graph.nodes()) == 3  # noqa: PLR2004
        assert len(graph.edges()) == 2  # noqa: PLR2004
        # Check topological order is possible
        try:
            list(nx.topological_sort(graph))
        except nx.NetworkXUnfeasible:
            pytest.fail("Graph is not a DAG")

    def test_parameter_without_dependencies(self) -> None:
        """Test parameter without any dependencies."""
        config = {
            "constant_param": {
                "sampler": {
                    "arguments": [],
                },
            },
        }
        graph = build_dependency_graph(config)
        assert "constant_param" in graph.nodes()
        assert len(graph.edges()) == 0

    def test_complex_real_world_example(self) -> None:
        """Test with a complex real-world-like config."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": "@alpha1",
                        "alpha_2": 4.51,
                        "transition": "@transition_mass",
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": "@peak1_mean",
                        "sigma_1": "@peak1_sigma",
                        "mean_2": "@peak2_mean",
                        "sigma_2": "@peak2_sigma",
                        "taper_range": "@taper",
                        "lambda_0": "@lambda0",
                        "lambda_1": "@lambda1",
                    },
                },
            },
            "mass_ratio": {
                "transform": "@mass_2 / @mass_1",
            },
            "mass_2": {
                "sampler": {
                    "arguments": ["@mass_1"],
                },
            },
            "redshift": {
                "sampler": {
                    "arguments": [],
                },
            },
            "luminosity_distance": {
                "transform": "@redshift * @scale_factor",
            },
            "combined_mass": {
                "transform": "@mass_1 + @mass_2",
            },
        }
        graph = build_dependency_graph(config)

        # Check all nodes
        expected_nodes = {
            "mass_1",
            "mass_ratio",
            "mass_2",
            "redshift",
            "luminosity_distance",
            "combined_mass",
            "alpha1",
            "transition_mass",
            "peak1_mean",
            "peak1_sigma",
            "peak2_mean",
            "peak2_sigma",
            "taper",
            "lambda0",
            "lambda1",
            "scale_factor",
        }
        assert set(graph.nodes()) == expected_nodes

        # Check key edges
        assert ("alpha1", "mass_1") in graph.edges()
        assert ("transition_mass", "mass_1") in graph.edges()
        assert ("mass_1", "mass_ratio") in graph.edges()
        assert ("mass_2", "mass_ratio") in graph.edges()
        assert ("redshift", "luminosity_distance") in graph.edges()
        assert ("scale_factor", "luminosity_distance") in graph.edges()
        assert ("mass_1", "combined_mass") in graph.edges()
        assert ("mass_2", "combined_mass") in graph.edges()



"""Broken power law plus two peaks distribution with Planck tapering."""

from __future__ import annotations

import jax
import jax.numpy as jnp
from jax import Array

from gwmock_pop.distributions.planck_tapered_broken_power_law_plus_two_peaks import (
    planck_tapered_broken_power_law_plus_two_peaks_cdf,
)


def planck_tapered_broken_power_law_plus_two_peaks(  # noqa: PLR0913
    key: Array,
    n_samples: int,
    alpha_1: float,
    alpha_2: float,
    transition: float,
    minimum: float,
    maximum: float,
    mean_1: float,
    sigma_1: float,
    mean_2: float,
    sigma_2: float,
    taper_range: float,
    lambda_0: float,
    lambda_1: float,
    n_grids: int = 1000,
) -> Array:
    """Draw samples from the Planck tapered broker power law plus two peaks distribution.

    Args:
        key: Random key.
        n_samples: Number of samples.
        alpha_1: Spectral index of the first power law.
        alpha_2: Spectral index of the second power law.
        transition: Transition point of the broken power law.
        minimum: Minimum of the distribution.
        maximum: Maximum of the distribution.
        mean_1: Mean of the first peak.
        sigma_1: Standard deviation of the first peak.
        mean_2: Mean of the second peak.
        sigma_2: Standard deviation of the second peak.
        taper_range: Range of the tapering.
        lambda_0: Mixing fraction of the power law.
        lambda_1: Mixing fraction of the first peak.
        n_grids: Number of grid points to evaluate the CDF.

    Returns:
        An array of samples.
    """
    # Compute the cdf
    x, cdf = planck_tapered_broken_power_law_plus_two_peaks_cdf(
        alpha_1=alpha_1,
        alpha_2=alpha_2,
        transition=transition,
        minimum=minimum,
        maximum=maximum,
        mean_1=mean_1,
        sigma_1=sigma_1,
        mean_2=mean_2,
        sigma_2=sigma_2,
        taper_range=taper_range,
        lambda_0=lambda_0,
        lambda_1=lambda_1,
        n_grids=n_grids,
    )

    # Draw uniform samples
    uniform_samples = jax.random.uniform(key=key, shape=(n_samples,))
    return jnp.interp(x=uniform_samples, xp=cdf, fp=x)



"""Functions to draw samples."""

from __future__ import annotations

from gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks import (
    planck_tapered_broken_power_law_plus_two_peaks,
)
from gwmock_pop.samplers.planck_tapered_conditional_ratio_power_law import planck_tapered_conditional_ratio_power_law

__all__ = ["planck_tapered_broken_power_law_plus_two_peaks", "planck_tapered_conditional_ratio_power_law"]



"""Conversion functions for parameters."""

from __future__ import annotations

from gwmock_pop.conversion.cbc import (
    compute_chirp_mass_from_mass_1_mass_2,
    compute_mass_ratio_from_mass_1_mass_2,
    compute_source_frame_mass_from_detector_frame_mass,
    compute_symmetric_mass_ratio_from_mass_1_mass_2,
    compute_total_mass_from_mass_1_mass_2,
)

__all__ = [
    "compute_chirp_mass_from_mass_1_mass_2",
    "compute_mass_ratio_from_mass_1_mass_2",
    "compute_source_frame_mass_from_detector_frame_mass",
    "compute_symmetric_mass_ratio_from_mass_1_mass_2",
    "compute_total_mass_from_mass_1_mass_2",
]



"""Conditional ratio power law sampler with Planck tapering."""

from __future__ import annotations

import jax
import jax.numpy as jnp
from jax import Array

from gwmock_pop.distributions.planck_tapered_conditional_ratio_power_law import (
    planck_tapered_conditional_ratio_power_law_cdf,
)


def planck_tapered_conditional_ratio_power_law(  # noqa: PLR0913
    key: Array,
    n_samples: int,
    denominator: Array,
    beta: float,
    numerator_minimum: float,
    taper_range: float,
    minimum: float,
    maximum: float,
    n_grids: int,
) -> Array:
    """Sample mass ratios with conditional power law and Planck tapering.

    Args:
        key: JAX random key.
        n_samples: Number of samples to generate.
        denominator: Denominator (scalar or array).
        beta: Spectral index of the ratio.
        numerator_minimum: Minimum of the numerator.
        taper_range: Range of the tapering.
        minimum: Minimum of the ratio.
        maximum: Maximum of the ratio.
        n_grids: Number of grid points to evaluate the CDF.

    Returns:
        Array of sampled mass ratios.
    """
    # Handle array denominators by generating samples for each element
    denominator_shape = jnp.shape(denominator)
    is_array = len(denominator_shape) > 0 and denominator_shape[0] > 0

    if is_array:
        # Denominator is an array - generate one sample per element
        n_denom = denominator_shape[0]
        keys = jax.random.split(key, n_denom)
        samples_list = []
        for _i, (d, k) in enumerate(zip(denominator, keys, strict=True)):
            x, cdf = planck_tapered_conditional_ratio_power_law_cdf(
                denominator=d,
                beta=beta,
                numerator_minimum=numerator_minimum,
                taper_range=taper_range,
                minimum=minimum,
                maximum=maximum,
                n_grids=n_grids,
            )
            uniform_sample = jax.random.uniform(key=k)
            sample = jnp.interp(x=uniform_sample, xp=cdf, fp=x)
            samples_list.append(sample)
        return jnp.array(samples_list)
    else:
        # Denominator is scalar - original behavior
        x, cdf = planck_tapered_conditional_ratio_power_law_cdf(
            denominator=denominator,
            beta=beta,
            numerator_minimum=numerator_minimum,
            taper_range=taper_range,
            minimum=minimum,
            maximum=maximum,
            n_grids=n_grids,
        )

        # Draw uniform samples
        uniform_samples = jax.random.uniform(key=key, shape=(n_samples,))
        return jnp.interp(x=uniform_samples, xp=cdf, fp=x)



"""Simulate populations."""

from __future__ import annotations

from pathlib import Path
from typing import Annotated, Any

import typer
from pydantic import ValidationError

from gwmock_pop.config.main import MainConfiguration
from gwmock_pop.simulators.graph import GraphSimulator


def _get_output_path(config: MainConfiguration) -> Path:
    """Build the output path for a simulation run."""
    output_dir = Path(config.run.output.directory)
    suffixes = {"csv": ".csv", "hdf5": ".hdf5", "npz": ".npz"}
    return output_dir / f"{config.run.name}{suffixes[config.run.output.format]}"


def _build_output_metadata(config: MainConfiguration, simulator: GraphSimulator) -> dict[str, Any]:
    """Build generic output metadata for formats that can persist it."""
    return {
        "config_version": config.config_version,
        "run_name": config.run.name,
        "n_samples": config.run.n_samples,
        "seed": config.run.seed,
        "parameter_names": simulator.parameter_names,
    }


def simulate_command(filename: Annotated[str, typer.Argument(help="File name of the configuration file.")]) -> None:
    """Simulate a population from a configuration file.

    Args:
        filename: File name of the configuration file.
    """
    import logging  # noqa: PLC0415

    logger = logging.getLogger("gwmock_pop")

    try:
        config = MainConfiguration.from_file(filename)
    except (FileNotFoundError, ValidationError, ValueError) as error:
        logger.error("Failed to load configuration from %s: %s", filename, error)
        raise typer.Exit(1) from error

    if config.run.mode != "fixed_n_samples":
        logger.error("Only run.mode='fixed_n_samples' is supported by the CLI MVP.")
        raise typer.Exit(1)

    if not config.parameters:
        logger.error("No parameter graph configuration found under 'parameters'.")
        raise typer.Exit(1)

    output_path = _get_output_path(config)
    output_path.parent.mkdir(parents=True, exist_ok=True)

    if output_path.exists() and not config.run.output.overwrite:
        logger.error(
            "Refusing to overwrite existing output file %s. Set run.output.overwrite=true to replace it.",
            output_path,
        )
        raise typer.Exit(1)

    simulator = GraphSimulator(config=config.parameters, seed=config.run.seed)
    population = simulator.simulate(n_samples=config.run.n_samples)

    metadata = None
    if config.run.output.save_metadata:
        metadata = _build_output_metadata(config, simulator)

    simulator.save(
        output_path=output_path,
        file_format=config.run.output.format,
        data=population,
        compression=config.run.output.compression,
        metadata=metadata,
    )
    logger.info("Saved %s samples to %s", population.shape[0], output_path)



"""Tests for flat lambda CDM cosmological functions."""

from __future__ import annotations

import jax.numpy as jnp
from jax import Array

from gwmock_pop.constants import SPEED_OF_LIGHT
from gwmock_pop.cosmology.flat_lambda_cdm import (
    compute_comoving_distance,
    compute_differential_comoving_volume,
    compute_hubble_parameter,
    compute_normalized_hubble_parameter,
)


class TestFlatLambdaCDM:
    """Test suite for flat lambda CDM cosmological functions."""

    def test_compute_normalized_hubble_parameter(self) -> None:
        """Test the normalized Hubble parameter computation."""
        # Test with a simple case
        redshift = jnp.array(1.0)
        omega_m = jnp.array(0.3)

        result = compute_normalized_hubble_parameter(redshift=redshift, omega_m=omega_m)

        # Expected: sqrt(0.3 * (1 + 1)^3 + (1 - 0.3)) = sqrt(0.3 * 8 + 0.7) = sqrt(2.4 + 0.7) = sqrt(3.1)
        expected = jnp.sqrt(0.3 * 8 + 0.7)

        assert isinstance(result, Array)
        assert jnp.isclose(result, expected, atol=1e-10)

    def test_compute_normalized_hubble_parameter_edge_cases(self) -> None:
        """Test normalized Hubble parameter with edge cases."""
        # Test with redshift = 0
        result = compute_normalized_hubble_parameter(redshift=jnp.array(0.0), omega_m=jnp.array(0.3))
        expected = jnp.sqrt(0.3 + (1 - 0.3))  # sqrt(0.3 + 0.7) = 1.0
        assert jnp.isclose(result, expected, atol=1e-10)

        # Test with omega_m = 1.0 (pure radiation)
        result = compute_normalized_hubble_parameter(redshift=jnp.array(1.0), omega_m=jnp.array(1.0))
        expected = jnp.sqrt(1.0 * (1.0 + 1.0) ** 3)
        assert jnp.isclose(result, expected, atol=1e-10)

    def test_compute_hubble_parameter(self) -> None:
        """Test the Hubble parameter computation."""
        redshift = jnp.array(1.0)
        hubble_constant = jnp.array(70.0)  # km/s/Mpc
        omega_m = jnp.array(0.3)

        result = compute_hubble_parameter(redshift=redshift, hubble_constant=hubble_constant, omega_m=omega_m)

        # Should equal H0 * normalized_hubble_parameter
        expected_normalized = compute_normalized_hubble_parameter(redshift=redshift, omega_m=omega_m)
        expected = hubble_constant * expected_normalized

        assert isinstance(result, Array)
        assert jnp.isclose(result, expected, atol=1e-10)

    def test_compute_comoving_distance(self) -> None:
        """Test the comoving distance computation."""
        redshift = jnp.array(1.0)
        hubble_constant = jnp.array(70.0)  # km/s/Mpc
        omega_m = jnp.array(0.3)

        result = compute_comoving_distance(
            redshift=redshift, hubble_constant=hubble_constant, omega_m=omega_m, n_grid=1000
        )

        assert isinstance(result, Array)
        # Should be positive
        assert result > 0.0

    def test_compute_comoving_distance_edge_cases(self) -> None:
        """Test comoving distance with edge cases."""
        # Test with redshift = 0
        result = compute_comoving_distance(
            redshift=jnp.array(0.0), hubble_constant=jnp.array(70.0), omega_m=jnp.array(0.3), n_grid=1000
        )
        assert isinstance(result, Array)
        assert result == 0.0

    def test_compute_differential_comoving_volume(self) -> None:
        """Test the differential comoving volume computation."""
        redshift = jnp.array(1.0)
        hubble_constant = jnp.array(70.0)  # km/s/Mpc
        omega_m = jnp.array(0.3)

        result = compute_differential_comoving_volume(
            redshift=redshift, hubble_constant=hubble_constant, omega_m=omega_m, n_grid=1000
        )

        assert isinstance(result, Array)
        # Should be positive for physical cases
        assert result > 0.0

    def test_compute_differential_comoving_volume_edge_cases(self) -> None:
        """Test differential comoving volume with edge cases."""
        # Test with redshift = 0
        result = compute_differential_comoving_volume(
            redshift=jnp.array(0.0), hubble_constant=jnp.array(70.0), omega_m=jnp.array(0.3), n_grid=1000
        )
        # Should be zero at z=0 (no volume)
        assert isinstance(result, Array)
        assert result >= 0.0

    def test_comprehensive_cosmological_functions(self) -> None:
        """Test comprehensive behavior of all cosmological functions."""
        redshift = jnp.array(2.0)
        hubble_constant = jnp.array(70.0)
        omega_m = jnp.array(0.3)

        # Test all functions work together
        normalized_hubble = compute_normalized_hubble_parameter(redshift=redshift, omega_m=omega_m)
        hubble = compute_hubble_parameter(redshift=redshift, hubble_constant=hubble_constant, omega_m=omega_m)
        comoving_distance = compute_comoving_distance(
            redshift=redshift, hubble_constant=hubble_constant, omega_m=omega_m, n_grid=1000
        )
        differential_volume = compute_differential_comoving_volume(
            redshift=redshift, hubble_constant=hubble_constant, omega_m=omega_m, n_grid=1000
        )

        # Verify types
        assert isinstance(normalized_hubble, Array)
        assert isinstance(hubble, Array)
        assert isinstance(comoving_distance, Array)
        assert isinstance(differential_volume, Array)

        # Verify reasonable ranges
        assert normalized_hubble > 0.0
        assert hubble > 0.0
        assert comoving_distance >= 0.0
        assert differential_volume >= 0.0

    def test_vectorized_inputs(self) -> None:
        """Test that functions work with vectorized inputs."""
        redshifts = jnp.array([0.0, 1.0, 2.0])
        omega_m = jnp.array(0.3)

        # Test normalized Hubble parameter with array input
        result = compute_normalized_hubble_parameter(redshift=redshifts, omega_m=omega_m)

        assert isinstance(result, Array)
        assert result.shape == (3,)

    def test_numerical_integration_accuracy(self) -> None:
        """Test that the numerical integration produces reasonable results."""
        # Test with known simple case: omega_m = 1.0 (Einstein-de Sitter)
        redshift = jnp.array(1.0)
        hubble_constant = jnp.array(70.0)
        omega_m = jnp.array(1.0)

        # Exact comoving distance for Einstein-de Sitter:
        # d_C = 2*c/H_0 * (1 - 1/sqrt(1+z))
        distance = compute_comoving_distance(
            redshift=redshift, hubble_constant=hubble_constant, omega_m=omega_m, n_grid=1000
        )
        expected = SPEED_OF_LIGHT / 1000 / hubble_constant * 2.0 * (1.0 - 1.0 / jnp.sqrt(1.0 + redshift))

        assert jnp.isclose(distance, expected, rtol=1e-3)

    def test_parameter_ranges(self) -> None:
        """Test that functions handle parameter ranges appropriately."""
        redshift = jnp.array(0.5)
        hubble_constant = jnp.array(50.0)  # Lower H0
        omega_m = jnp.array(0.5)  # Mid-range matter density

        result = compute_comoving_distance(
            redshift=redshift, hubble_constant=hubble_constant, omega_m=omega_m, n_grid=1000
        )

        assert isinstance(result, Array)



"""Broken power law plus two peaks distribution with Planck tapering."""

from __future__ import annotations

import jax.numpy as jnp
from jax import Array
from jax.experimental import checkify
from jax.nn import logsumexp
from jax.scipy.stats.norm import logpdf as norm_logpdf

from gwmock_pop.distributions.broken_power_law import broken_power_law_unnormalized_logpdf
from gwmock_pop.distributions.smoothing import log_planck_tapering_function
from gwmock_pop.integrators.trapezoid import log_trapezoidal_cumsum


def planck_tapered_broken_power_law_plus_two_peaks_unnormalized_logpdf(  # noqa: PLR0913
    x: Array,
    alpha_1: float,
    alpha_2: float,
    transition: float,
    minimum: float,
    maximum: float,
    mean_1: float,
    sigma_1: float,
    mean_2: float,
    sigma_2: float,
    taper_range: float,
    lambda_0: float,
    lambda_1: float,
) -> Array:
    """Compute the unnormalized logpdf of Planck tapered broken power law plus two peaks distribution.

    Args:
        x: An array of values to evaluate the unnormalized pdf.
        alpha_1: Spectral index of the first power law.
        alpha_2: Spectral index of the second power law.
        transition: Transition point of the broken power law.
        minimum: Minimum of the distribution.
        maximum: Maximum of the distribution.
        mean_1: Mean of the first peak.
        sigma_1: Standard deviation of the first peak.
        mean_2: Mean of the second peak.
        sigma_2: Standard deviation of the second peak.
        taper_range: Range of the tapering.
        lambda_0: Mixing fraction of the power law.
        lambda_1: Mixing fraction of the first peak.

    Returns:
        Unnormalized logpdf.
    """
    # Evaluate the logpdf of the broken power low
    broken_power_law_logpdf = broken_power_law_unnormalized_logpdf(
        x=x, x_min=minimum, x_max=maximum, transition=transition, alpha_1=alpha_1, alpha_2=alpha_2
    )

    # Evaluate the logpdf of the two normal peaks
    first_peak_logpdf = norm_logpdf(x=x, loc=mean_1, scale=sigma_1)
    second_peak_logpdf = norm_logpdf(x=x, loc=mean_2, scale=sigma_2)

    # Tapering function
    log_smoothing = log_planck_tapering_function(x=x, x_min=minimum, delta=taper_range)

    # Weights
    checkify.check(lambda_0 >= 0.0, "Invalid mixture weights: require lambda_0 >=0.")
    checkify.check(lambda_1 >= 0.0, "Invalid mixture weights: require lambda_1 >=0.")
    checkify.check(lambda_0 + lambda_1 <= 1, "Invalid mixture weights: require lambda_0 + lambda_1 <= 1.")
    weights = jnp.array([lambda_0, lambda_1, 1.0 - lambda_0 - lambda_1])

    # Stacked logpdf of the components
    stacked_logpdf = jnp.stack([broken_power_law_logpdf, first_peak_logpdf, second_peak_logpdf])

    return logsumexp(a=stacked_logpdf, b=weights[:, None], axis=0) + log_smoothing


def planck_tapered_broken_power_law_plus_two_peaks_cdf(  # noqa: PLR0913
    alpha_1: float,
    alpha_2: float,
    transition: float,
    minimum: float,
    maximum: float,
    mean_1: float,
    sigma_1: float,
    mean_2: float,
    sigma_2: float,
    taper_range: float,
    lambda_0: float,
    lambda_1: float,
    n_grids: int,
) -> tuple[Array, Array]:
    """Compute the cdf of Planck tapered broken power law plus two peaks distribution.

    Args:
        alpha_1: Spectral index of the first power law.
        alpha_2: Spectral index of the second power law.
        transition: Transition point of the broken power law.
        minimum: Minimum of the distribution.
        maximum: Maximum of the distribution.
        mean_1: Mean of the first peak.
        sigma_1: Standard deviation of the first peak.
        mean_2: Mean of the second peak.
        sigma_2: Standard deviation of the second peak.
        taper_range: Range of the tapering.
        lambda_0: Mixing fraction of the power law.
        lambda_1: Mixing fraction of the first peak.
        n_grids: Number of grid points to evaluate the CDF.

    Returns:
        x: Grid points to evaluate the CDF.
        cdf: Cumulative distribution function.
    """
    # Create an array of grid points to evaluate the CDF
    x = jnp.linspace(start=minimum, stop=maximum, num=n_grids)

    # Compute the unnormalized logpdf
    unnormalized_logpdf = planck_tapered_broken_power_law_plus_two_peaks_unnormalized_logpdf(
        x=x,
        alpha_1=alpha_1,
        alpha_2=alpha_2,
        transition=transition,
        minimum=minimum,
        maximum=maximum,
        mean_1=mean_1,
        sigma_1=sigma_1,
        mean_2=mean_2,
        sigma_2=sigma_2,
        taper_range=taper_range,
        lambda_0=lambda_0,
        lambda_1=lambda_1,
    )

    # Compute the logcdf
    logcdf = log_trapezoidal_cumsum(log_y=unnormalized_logpdf, x=x)

    # Compute the cdf
    return x, jnp.exp(logcdf - logcdf[-1])



"""Base class of the simulators."""

from __future__ import annotations

import json
from abc import ABC, abstractmethod
from collections.abc import Callable
from functools import wraps
from pathlib import Path
from typing import Any

import h5py
import jax.numpy as jnp
import networkx as nx
import numpy as np
from jax import Array


class Simulator(ABC):
    """Abstract base class for generating simulated populations."""

    def __init__(self, *args: object, **kwargs: object) -> None:
        """Initialize the instance.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.
        """
        self.graph = nx.DiGraph()
        self._last_data: Array | None = None
        self._node_funcs: dict[str, Callable] = {}
        self._node_depends: dict[str, list[str]] = {}

    @property
    @abstractmethod
    def parameter_names(self) -> list[str]:
        """Get the names of the parameters.

        Returns:
            List of parameter names.

        """

    def register_node(self, name: str, func: Callable, depends_on: list[str] | None = None) -> None:
        """Register a node function on this instance.

        Args:
            name: Parameter name.
            func: A function to simulate this parameter.
            depends_on: A list of dependent parameters.
        """
        dependencies = list(depends_on or [])
        if name in self.graph:
            self.graph.remove_edges_from((dep, name) for dep in list(self.graph.predecessors(name)))
        self._node_funcs[name] = func
        self._node_depends[name] = dependencies
        self.graph.add_node(name, func=func)
        for dep in dependencies:
            self.graph.add_edge(dep, name)

    def node(self, depends_on: list[str] | None = None) -> Callable:
        """Implement a decorator to bind a node to this instance.

        Args:
            depends_on: A list of dependencies.

        Returns:
            A callable.
        """

        def decorator(func: Callable):
            node_name = func.__name__  # ty:ignore[unresolved-attribute]
            self.register_node(node_name, func, depends_on)

            @wraps(func)
            def wrapper(*args: object, **kwargs: object):
                return func(*args, **kwargs)

            return wrapper

        return decorator

    @abstractmethod
    def _simulate_impl(self, *args: object, **kwargs: object) -> Array:
        """Implement simulation for subclass.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            2D array of shape (n_samples, n_parameters).

        """

    def simulate(self, *args: object, **kwargs: object) -> Array:
        """Simulate a population of sources.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            2D array of shape (n_samples, n_parameters).

        """
        result = self._simulate_impl(*args, **kwargs)
        self._validate_output(result)
        self._last_data = result
        return result

    def __call__(self, *args: object, **kwargs: object) -> Array:
        """Call simulate() with n_samples.

        Args:
            *args: Positional arguments.
            **kwargs: Keyword arguments.

        Returns:
            2D array of shape (n_samples, n_parameters).

        """
        return self.simulate(*args, **kwargs)

    def save(
        self,
        output_path: str | Path,
        file_format: str | None = None,
        data: Array | None = None,
        compression: str | None = None,
        metadata: dict[str, Any] | None = None,
    ) -> None:
        """Save the simulated data to a file.

        Args:
            output_path: Path to save the file.
            file_format: File format (npy, npz, csv, hdf5). If None, infers from extension.
            data: Data to save. If None, uses last simulated data.
            compression: Optional compression setting for supported formats.
                For HDF5, this specifies the compression filter (e.g., "gzip").
                For NPZ, any non-None value enables default zlib compression.
            metadata: Optional metadata to store alongside the samples.

        """
        output_path = Path(output_path)

        if file_format is None:
            file_format = output_path.suffix.lstrip(".")

        if data is None:
            if self._last_data is None:
                raise ValueError("No data provided and no last simulated data available.")
            data = self._last_data

        if file_format == "npy":
            jnp.save(output_path, data)
        elif file_format == "npz":
            if metadata:
                payload = {"data": np.asarray(data), "__metadata__": np.array(json.dumps(metadata))}
            else:
                payload = {"data": np.asarray(data)}
            if compression is None:
                np.savez(output_path, **payload)
            else:
                np.savez_compressed(output_path, **payload)
        elif file_format == "csv":
            np.savetxt(output_path, data, delimiter=",")
        elif file_format == "hdf5":
            self._save_hdf5(output_path, data, compression=compression, metadata=metadata)
        else:
            raise ValueError(f"Unsupported format: {file_format}")

    def load(self, input_path: str | Path) -> Array:
        """Load data from a file.

        Args:
            input_path: Path to load the file from.

        Returns:
            Loaded data as numpy array.

        """
        input_path = Path(input_path)

        suffix = input_path.suffix.lstrip(".")

        if suffix == "npy":
            return jnp.load(input_path)
        elif suffix == "npz":
            with np.load(input_path) as loaded:
                return jnp.array(loaded["data"])
        elif suffix == "csv":
            return jnp.array(np.loadtxt(input_path, delimiter=","))
        elif suffix == "hdf5":
            return self._load_hdf5(input_path)
        else:
            raise ValueError(f"Unsupported format: {suffix}")

    @staticmethod
    def _save_hdf5(
        path: Path, data: Array, compression: str | None = None, metadata: dict[str, Any] | None = None
    ) -> None:
        """Save data to HDF5 format.

        Args:
            path: Path to save the file.
            data: Data to save.
            compression: Optional HDF5 compression filter.
            metadata: Optional metadata to store in the file.

        """
        with h5py.File(path, "w") as f:
            f.create_dataset("data", data=data, compression=compression)
            if metadata:
                metadata_group = f.create_group("metadata")
                for key, value in metadata.items():
                    metadata_group.attrs[key] = json.dumps(value)

    @staticmethod
    def _load_hdf5(path: Path) -> Array:
        """Load data from HDF5 format.

        Args:
            path: Path to load the file from.

        Returns:
            Loaded data as numpy array.

        """
        with h5py.File(path, "r") as f:
            data_obj = f["data"]
            if not isinstance(data_obj, h5py.Dataset):
                raise TypeError('f["data"] is not a h5py.Dataset instance.')
            return jnp.array(data_obj[()])

    def _validate_output(self, array: Array) -> None:
        """Validate the output of simulate().

        Args:
            array: Array to validate.

        Raises:
            ValueError: If array shape is invalid.

        """
        expected_ndim = 2
        if array.ndim != expected_ndim:
            raise ValueError(f"Expected 2D array, got {array.ndim}D array")

        expected_n_parameters = len(self.parameter_names)
        if array.shape[1] != expected_n_parameters:
            raise ValueError(f"Expected {expected_n_parameters} parameters, got {array.shape[1]}")



"""Graph-based population simulator."""

from __future__ import annotations

import secrets
from pathlib import Path
from typing import Any

import jax
import jax.numpy as jnp
import networkx as nx
from jax import Array

from gwmock_pop.graph.build import build_dependency_graph
from gwmock_pop.mixins.random import RandomMixin
from gwmock_pop.simulators.simulator import Simulator
from gwmock_pop.utils.import_utils import import_from_string
from gwmock_pop.utils.yaml import read_data_file


class GraphSimulator(RandomMixin, Simulator):
    """Graph-based population simulator.

    This simulator uses a probabilistic graphical model to generate populations.
    Parameters are defined in a configuration with dependencies, and the simulator
    executes sampling in topological order based on the dependency graph.

    Args:
        config: Configuration dictionary defining parameters and their sampling/transform rules.
        **kwargs: Additional arguments passed to parent class.

    Example:
        >>> config = {
        ...     "mass_1": {
        ...         "sampler": {
        ...             "function": "planck_tapered_broken_power_law_plus_two_peaks",
        ...             "arguments": {
        ...                 "alpha_1": 1.72,
        ...                 "alpha_2": 4.51,
        ...                 "transition": 35.6,
        ...                 "minimum": 5.06,
        ...                 "maximum": 300.0,
        ...             },
        ...         },
        ...     },
        ...     "mass_ratio": {
        ...         "sampler": {
        ...             "function": "planck_tapered_conditional_ratio_power_law",
        ...             "arguments": {"denominator": "@mass_1"},
        ...         },
        ...     },
        ... }
        >>> simulator = GraphSimulator(config=config)
        >>> population = simulator()
    """

    def __init__(
        self,
        config: dict[str, Any],
        source_type: str | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the graph-based simulator.

        Args:
            config: Configuration dictionary with parameter definitions.
            source_type: Optional logical source identifier for higher-level orchestration.
            **kwargs: Additional arguments passed to parent class.
        """
        super().__init__(**kwargs)
        self._config = config
        self._source_type = source_type
        self._sampled_values: dict[str, Any] = {}
        self._parameter_names = [name for name, spec in config.items() if self._include_in_output(spec)]
        self._build_graph()

    @property
    def parameter_names(self) -> list[str]:
        """Get the names of the parameters.

        Returns:
            List of parameter names.
        """
        return self._parameter_names

    @property
    def source_type(self) -> str | None:
        """Get the optional logical source type.

        Returns:
            Source type string, when configured.
        """
        return self._source_type

    @staticmethod
    def _include_in_output(spec: dict[str, Any]) -> bool:
        """Determine whether a parameter should be included in the final output."""
        return not bool(spec.get("exclude", False) or spec.get("intermediate", False))

    def _build_graph(self) -> None:
        """Build the dependency graph from the configuration."""
        # Extract only parameters with samplers (not transforms)
        self._graph = build_dependency_graph(self._config)

    def _get_ordered_parameters(self) -> list[str]:
        """Get parameters in topological order.

        Returns:
            List of parameter names in sampling order.
        """
        ordered = list(nx.topological_sort(self._graph))
        undefined = [name for name in ordered if name not in self._config]
        if undefined:
            raise ValueError(f"Undefined parameter dependencies: {undefined}")
        return ordered

    def _execute_sampler(
        self,
        sampler_name: str,
        arguments: dict[str, Any],
        n_samples: int | None = None,
    ) -> Array:
        """Execute a sampler with given arguments.

        Args:
            sampler_name: Name of the sampler function.
            arguments: Arguments for the sampler.
            n_samples: Number of samples to inject when the sampler arguments omit it.

        Returns:
            Sampled array.
        """
        # Resolve dependencies in arguments
        if arguments is None:
            arguments = {}
        elif not isinstance(arguments, dict):
            raise ValueError("Sampler arguments must be a mapping (or null).")
        resolved_args: dict[str, Any] = {}
        for key, value in arguments.items():
            if isinstance(value, str) and value.startswith("@"):
                dep_name = value[1:]  # Remove @ prefix
                if dep_name in self._sampled_values:
                    resolved_args[key] = self._sampled_values[dep_name]
                else:
                    raise ValueError(f"Dependency '{dep_name}' not sampled yet")
            else:
                resolved_args[key] = value

        if n_samples is not None and "n_samples" not in resolved_args:
            resolved_args["n_samples"] = n_samples

        # Add random key if needed
        if "key" not in resolved_args:
            resolved_args["key"] = self.rng_manager.new_key

        # Dispatch to sampler function
        sampler_func = import_from_string(object_path=sampler_name, default_module="gwmock_pop.samplers")
        return sampler_func(**resolved_args)

    def _execute_transform(
        self,
        transform_spec: dict[str, Any],
    ) -> Array:
        """Execute a transform.

        Args:
            transform_spec: Transform specification.

        Returns:
            Transformed array.
        """
        if isinstance(transform_spec, str):
            raise ValueError("String transform expressions are not supported by GraphSimulator yet.")

        function_name = transform_spec.get("function", "")
        arguments = transform_spec.get("arguments")
        if arguments is None:
            arguments = {}
        elif not isinstance(arguments, dict):
            raise ValueError("Only mapping-style transform arguments are currently supported.")

        # Resolve dependencies
        resolved_args: dict[str, Any] = {}
        for key, value in arguments.items():
            if isinstance(value, str) and value.startswith("@"):
                dep_name = value[1:]
                if dep_name in self._sampled_values:
                    resolved_args[key] = self._sampled_values[dep_name]
                else:
                    raise ValueError(f"Dependency '{dep_name}' not available for transform")
            else:
                resolved_args[key] = value

        # Dispatch to transform function
        transform_func = import_from_string(object_path=function_name, default_module="gwmock_pop.transforms")
        return transform_func(**resolved_args)

    def _coerce_output_column(
        self,
        parameter_name: str,
        value: Any,
        expected_n_samples: int | None,
    ) -> tuple[Array, int]:
        """Convert a sampled value into a validated output column."""
        array = jnp.asarray(value)
        if array.ndim == 0:
            raise ValueError(f"Parameter '{parameter_name}' produced a scalar output, expected an array of samples.")

        current_n_samples = int(array.shape[0])
        if expected_n_samples is not None and current_n_samples != expected_n_samples:
            raise ValueError(
                f"Parameter '{parameter_name}' produced {current_n_samples} samples, expected {expected_n_samples}."
            )
        return array, current_n_samples

    def _simulate_impl(self, n_samples: int | None = None, **kwargs: Any) -> Array:
        """Implement simulation using graph traversal.

        Args:
            n_samples: Number of samples to inject into samplers when not explicitly configured.
            **kwargs: Keyword arguments.

        Returns:
            2D array of shape (n_samples, n_parameters).
        """
        del kwargs

        self._sampled_values = {}

        # Get ordered parameters
        ordered_params = self._get_ordered_parameters()

        # Sample each parameter in order
        for param_name in ordered_params:
            spec = self._config[param_name]

            if "sampler" in spec:
                sampler_spec = spec["sampler"]
                sampler_name = sampler_spec.get("function", "")
                sampler_args = sampler_spec.get("arguments", {})
                samples = self._execute_sampler(sampler_name, sampler_args, n_samples=n_samples)
                self._sampled_values[param_name] = samples

            elif "transform" in spec:
                transform_spec = spec["transform"]
                transformed = self._execute_transform(transform_spec)
                self._sampled_values[param_name] = transformed

        if not self.parameter_names:
            raise ValueError("GraphSimulator configuration does not define any output parameters.")

        expected_n_samples = n_samples
        output_columns: list[Array] = []
        for parameter_name in self.parameter_names:
            output_column, expected_n_samples = self._coerce_output_column(
                parameter_name=parameter_name,
                value=self._sampled_values[parameter_name],
                expected_n_samples=expected_n_samples,
            )
            output_columns.append(output_column)

        # Build output array in parameter order
        output = jnp.column_stack(output_columns)
        return output

    @classmethod
    def from_config_file(cls, config_path: str | Path, encoding: str = "utf-8", **kwargs: Any) -> GraphSimulator:
        """Create simulator from configuration file.

        Args:
            config_path: Path to YAML/TOML configuration file.
            encoding: Encoding of the file.
            **kwargs: Additional arguments passed to __init__.

        Returns:
            Configured simulator instance.
        """
        config_path = Path(config_path)
        try:
            config = read_data_file(config_path, encoding=encoding)
        except ValueError as error:
            message = str(error)
            if "Suffix of filename=" in message:
                raise ValueError(
                    f"Suffix of config_path={config_path} is not supported. Only '.yaml', '.yml', and '.toml' are supported."
                ) from error
            if "mapping at the top level" in message:
                raise ValueError("Config file must contain a mapping at the root.") from error
            raise

        # Extract parameters section if present
        if "parameters" in config:
            config = config["parameters"]

        return cls(config=config, **kwargs)

    def reset(self) -> None:
        """Reset the simulator state."""
        self._sampled_values = {}
        # Reset RNG by using a fresh key
        if hasattr(self, "_rng_manager"):
            seed = getattr(self._rng_manager, "_seed", None)
            if seed is not None:
                self._rng_manager.key = jax.random.key(seed)
            else:
                self._rng_manager.key = jax.random.key(secrets.randbelow(2**63))



"""Functions to build a dependency graph."""

from __future__ import annotations

from typing import Any

from networkx import DiGraph

from gwmock_pop.graph.generic import extract_dependencies_from_spec, extract_references
from gwmock_pop.graph.sampler import extract_sampler_dependencies
from gwmock_pop.graph.transform import extract_transform_dependencies


def build_dependency_graph(parameters_config: dict[str, Any]) -> DiGraph:
    """Build a dependency graph.

    Build a directed graph here:

    - nodes = parameter names
    - edge A → B means: "to compute/sample B, you need A first"

    Args:
        parameters_config: A dictionary of parameters.

    Returns:
        A DiGraph for topological sort.
    """
    graph = DiGraph()

    for parameter_name, spec in parameters_config.items():
        # Add all parameters to the graph
        graph.add_node(node_for_adding=parameter_name)

        # Case 1: Sampler block
        if "sampler" in spec:
            dependencies = extract_sampler_dependencies(sampler_spec=spec["sampler"])
            add_dependencies_to_graph(graph=graph, dependencies=dependencies, parameter_name=parameter_name)

        # Case 1: simple transform expression like "@mass_1 * @mass_ratio"
        if "transform" in spec:
            dependencies = extract_transform_dependencies(transform=spec["transform"])
            add_dependencies_to_graph(graph=graph, dependencies=dependencies, parameter_name=parameter_name)

        # Case 2: condition Expression
        if "condition" in spec:
            dependencies = extract_references(expr=spec["condition"])
            add_dependencies_to_graph(graph=graph, dependencies=dependencies, parameter_name=parameter_name)

        # Case 3: condition_else block (recursive)
        if "condition_else" in spec:
            dependencies = extract_dependencies_from_spec(spec=spec["condition_else"])
            add_dependencies_to_graph(graph=graph, dependencies=dependencies, parameter_name=parameter_name)

        # 4. branches style
        if "branches" in spec:
            for branch in spec["branches"]:
                # condition in branch
                if "condition" in branch:
                    dependencies = extract_references(expr=branch["condition"])
                    add_dependencies_to_graph(graph=graph, dependencies=dependencies, parameter_name=parameter_name)
                # recursive into branch content
                branch_dependencies = extract_dependencies_from_spec(spec=branch)
                add_dependencies_to_graph(graph=graph, dependencies=branch_dependencies, parameter_name=parameter_name)

    return graph


def add_dependencies_to_graph(graph: DiGraph, dependencies: set[str], parameter_name: str) -> None:
    """Add dependencies of a parameter to the graph.

    Args:
        graph: Dependency graph.
        dependencies: A set of dependencies.
        parameter_name: Parameter name.
    """
    for dependency in dependencies:
        graph.add_edge(u_of_edge=dependency, v_of_edge=parameter_name)



"""Functions to extract dependencies of transform."""

from __future__ import annotations

from typing import Any

from gwmock_pop.graph.generic import extract_references


def extract_transform_dependencies(transform: str | dict[str, Any]) -> set[str]:
    """Extract the dependencies of transform.

    Args:
        transform: Expression or dictionary of transform.

    Returns:
        Dependencies of transform.
    """
    if isinstance(transform, str):
        dependencies = extract_references(transform)
    elif isinstance(transform, dict) and "function" in transform:
        args = transform.get("arguments", [])

        if isinstance(args, list):
            dependencies = {arg[1:] for arg in args if isinstance(arg, str) and arg.startswith("@") and len(arg) > 1}
        elif isinstance(args, dict):
            dependencies = {v[1:] for v in args.values() if isinstance(v, str) and v.startswith("@") and len(v) > 1}
        else:
            dependencies = set()

    else:
        dependencies = set()

    return dependencies



"""Tests for GraphSimulator."""

from __future__ import annotations

import tempfile
from pathlib import Path
from typing import Any

import jax
import jax.numpy as jnp
import pytest
import yaml

import gwmock_pop.simulators.graph as graph_module
from gwmock_pop.simulators.graph import GraphSimulator

# Constants for test data
DEFAULT_N_SAMPLES = 100
DEFAULT_N_PARAMETERS = 2
DEFAULT_N_PARAMETERS_MULTI = 3


class TestGraphSimulator:
    """Test suite for GraphSimulator class."""

    def test_init_with_config(self) -> None:
        """Test initialization with a simple config."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        assert simulator is not None
        assert simulator.parameter_names == ["mass_1"]

    def test_init_with_multiple_parameters(self) -> None:
        """Test initialization with multiple parameters."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
            "mass_ratio": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_conditional_ratio_power_law",
                    "arguments": {
                        "denominator": "@mass_1",
                        "beta": 1.5,
                        "numerator_minimum": 0.5,
                        "taper_range": 0.5,
                        "minimum": 0.1,
                        "maximum": 1.0,
                        "n_grids": 100,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        assert simulator.parameter_names == ["mass_1", "mass_ratio"]

    def test_graph_nodes(self) -> None:
        """Test that graph nodes are correctly built."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
            "mass_ratio": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_conditional_ratio_power_law",
                    "arguments": {
                        "denominator": "@mass_1",
                        "beta": 1.5,
                        "numerator_minimum": 0.5,
                        "taper_range": 0.5,
                        "minimum": 0.1,
                        "maximum": 1.0,
                        "n_grids": 100,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        graph_nodes = list(simulator._graph.nodes())
        assert "mass_1" in graph_nodes
        assert "mass_ratio" in graph_nodes

    def test_source_type_defaults_to_none(self) -> None:
        """Test that source_type is optional."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        assert simulator.source_type is None

    def test_source_type_can_be_configured(self) -> None:
        """Test that source_type can be provided by the caller."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config, source_type="generic-source")
        assert simulator.source_type == "generic-source"

    def test_reset(self) -> None:
        """Test that reset clears sampled values."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        _ = simulator()

        assert len(simulator._sampled_values) > 0

        simulator.reset()

        assert len(simulator._sampled_values) == 0

    def test_from_config_file_yaml(self) -> None:
        """Test loading config from YAML file."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            yaml.dump(config, f)
            temp_path = f.name

        try:
            simulator = GraphSimulator.from_config_file(temp_path)
            assert simulator is not None
            assert simulator.parameter_names == ["mass_1"]
        finally:
            Path(temp_path).unlink()

    def test_from_config_file_toml(self) -> None:
        """Test loading config from TOML file."""
        config = """
[parameters.mass_1.sampler]
function = "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks"

[parameters.mass_1.sampler.arguments]
alpha_1 = 1.72
alpha_2 = 4.51
transition = 35.6
minimum = 5.06
maximum = 300.0
mean_1 = 9.76
sigma_1 = 0.649
mean_2 = 32.8
sigma_2 = 3.92
taper_range = 4.32
lambda_0 = 0.361
lambda_1 = 0.586
"""

        with tempfile.NamedTemporaryFile(mode="w", suffix=".toml", delete=False) as f:
            f.write(config)
            temp_path = f.name

        try:
            simulator = GraphSimulator.from_config_file(temp_path)
            assert simulator is not None
            assert simulator.parameter_names == ["mass_1"]
        finally:
            Path(temp_path).unlink()

    def test_from_config_file_unsupported_format(self) -> None:
        """Test that unsupported file format raises ValueError."""
        with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
            f.write("{}")
            temp_path = f.name

        try:
            with pytest.raises(ValueError, match="Suffix of config_path"):
                GraphSimulator.from_config_file(temp_path)
        finally:
            Path(temp_path).unlink()

    def test_from_config_file_with_parameters_section(self) -> None:
        """Test loading config with parameters section."""
        config = {
            "parameters": {
                "mass_1": {
                    "sampler": {
                        "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                        "arguments": {
                            "alpha_1": 1.72,
                            "alpha_2": 4.51,
                            "transition": 35.6,
                            "minimum": 5.06,
                            "maximum": 300.0,
                            "mean_1": 9.76,
                            "sigma_1": 0.649,
                            "mean_2": 32.8,
                            "sigma_2": 3.92,
                            "taper_range": 4.32,
                            "lambda_0": 0.361,
                            "lambda_1": 0.586,
                        },
                    },
                },
            },
        }

        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            yaml.dump(config, f)
            temp_path = f.name

        try:
            simulator = GraphSimulator.from_config_file(temp_path)
            assert simulator is not None
            assert simulator.parameter_names == ["mass_1"]
        finally:
            Path(temp_path).unlink()

    def test_init_with_custom_n_samples(self) -> None:
        """Test initialization with custom n_samples passed to sampler arguments."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                        "n_samples": 50,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        assert simulator is not None

        result = simulator()
        assert result.shape[0] == 50  # noqa: PLR2004
        assert result.shape[1] == len(simulator.parameter_names)

    def test_simulate_injects_n_samples_when_not_in_sampler_arguments(self) -> None:
        """Test that simulate(n_samples=...) injects the sample count into samplers."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        result = simulator.simulate(n_samples=12)

        assert result.shape == (12, 1)

    def test_simulate_with_multiple_parameters(self) -> None:
        """Test simulation with multiple parameters and dependencies."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                        "n_samples": 20,
                    },
                },
            },
            "mass_ratio": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_conditional_ratio_power_law",
                    "arguments": {
                        "denominator": "@mass_1",
                        "beta": 1.5,
                        "numerator_minimum": 0.5,
                        "taper_range": 0.5,
                        "minimum": 0.1,
                        "maximum": 1.0,
                        "n_grids": 100,
                        "n_samples": 20,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        result = simulator()

        assert result.shape[0] == 20  # noqa: PLR2004  # n_samples
        assert result.shape[1] == DEFAULT_N_PARAMETERS  # mass_1 and mass_ratio
        assert "mass_1" in simulator.parameter_names
        assert "mass_ratio" in simulator.parameter_names

    def test_output_shape_validation(self) -> None:
        """Test that output shape matches expected dimensions."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                        "n_samples": 75,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        result = simulator()

        assert result.ndim == 2  # noqa: PLR2004
        assert result.shape[0] == 75  # noqa: PLR2004
        assert result.shape[1] == len(simulator.parameter_names)

    def test_reset_with_custom_seed(self) -> None:
        """Test that reset preserves custom seed behavior."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config, seed=123)
        result1 = simulator()
        simulator.reset()
        result2 = simulator()

        # Results should be identical after reset with same seed
        assert jnp.allclose(result1, result2)

    def test_simulator_callable_without_args(self) -> None:
        """Test that simulator can be called with positional arguments."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        # simulator() should work without arguments
        result = simulator()
        assert result is not None
        assert result.shape[0] == DEFAULT_N_SAMPLES  # Default n_samples

    def test_parameter_names_from_config(self) -> None:
        """Test that parameter_names matches config keys."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
            "mass_ratio": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_conditional_ratio_power_law",
                    "arguments": {
                        "denominator": "@mass_1",
                        "beta": 1.5,
                        "numerator_minimum": 0.5,
                        "taper_range": 0.5,
                        "minimum": 0.1,
                        "maximum": 1.0,
                        "n_grids": 100,
                    },
                },
            },
            "redshift": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.0,
                        "alpha_2": 2.0,
                        "transition": 10.0,
                        "minimum": 0.01,
                        "maximum": 2.0,
                        "mean_1": 0.5,
                        "sigma_1": 0.1,
                        "mean_2": 1.0,
                        "sigma_2": 0.2,
                        "taper_range": 0.5,
                        "lambda_0": 0.5,
                        "lambda_1": 0.5,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        assert set(simulator.parameter_names) == set(config.keys())
        assert len(simulator.parameter_names) == DEFAULT_N_PARAMETERS_MULTI

    def test_excluded_parameters_are_not_returned(self) -> None:
        """Test that parameters marked as excluded are omitted from the output."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 10,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
            "hidden_mass": {
                "exclude": True,
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 10,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        result = simulator()

        assert simulator.parameter_names == ["mass_1"]
        assert result.shape == (10, 1)

    def test_source_type_is_stable(self) -> None:
        """Test that source_type remains stable across property access."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config, source_type="population")
        assert simulator.source_type == "population"

        # Test that it's consistent across multiple calls
        assert simulator.source_type == "population"
        assert simulator.source_type == "population"

    def test_graph_nodes_match_parameters(self) -> None:
        """Test that graph nodes match parameter names."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
            "mass_ratio": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_conditional_ratio_power_law",
                    "arguments": {"denominator": "@mass_1"},
                },
            },
        }

        simulator = GraphSimulator(config=config)
        graph_nodes = list(simulator._graph.nodes())

        # All parameter names should be in graph nodes
        for param_name in simulator.parameter_names:
            assert param_name in graph_nodes

        # All graph nodes should be parameter names
        for node in graph_nodes:
            assert node in simulator.parameter_names

    def test_sampled_values_after_simulation(self) -> None:
        """Test that sampled_values is populated after simulation."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                        "n_samples": 30,
                    },
                },
            },
            "mass_ratio": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_conditional_ratio_power_law",
                    "arguments": {
                        "denominator": "@mass_1",
                        "beta": 1.5,
                        "numerator_minimum": 0.5,
                        "taper_range": 0.5,
                        "minimum": 0.1,
                        "maximum": 1.0,
                        "n_grids": 100,
                        "n_samples": 30,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        _ = simulator()

        # Check that sampled_values is populated
        assert len(simulator._sampled_values) == 2  # noqa: PLR2004
        assert "mass_1" in simulator._sampled_values
        assert "mass_ratio" in simulator._sampled_values

        # Check that values are JAX arrays
        assert hasattr(simulator._sampled_values["mass_1"], "shape")
        assert hasattr(simulator._sampled_values["mass_ratio"], "shape")

        # Check shapes
        assert simulator._sampled_values["mass_1"].shape[0] == 30  # noqa: PLR2004
        assert simulator._sampled_values["mass_ratio"].shape[0] == 30  # noqa: PLR2004

    def test_dependency_not_sampled_raises_error(self) -> None:
        """Test that missing dependency raises ValueError."""
        config = {
            "mass_ratio": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_conditional_ratio_power_law",
                    "arguments": {
                        "denominator": "@mass_1",  # mass_1 not defined
                        "beta": 1.5,
                        "numerator_minimum": 0.5,
                        "taper_range": 0.5,
                        "minimum": 0.1,
                        "maximum": 1.0,
                        "n_grids": 100,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        with pytest.raises(ValueError, match="Undefined parameter dependencies"):
            _ = simulator()

    def test_sampler_with_explicit_key(self) -> None:
        """Test that sampler works with explicit key in arguments."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                        "key": jax.random.key(42),  # Explicit key
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        result = simulator()
        assert result.shape[0] == 100  # noqa: PLR2004  # default n_samples

    def test_sampler_with_explicit_n_samples(self) -> None:
        """Test that sampler works with explicit n_samples in arguments."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                        "n_samples": 15,  # Explicit n_samples
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        result = simulator()
        assert result.shape[0] == 15  # Should use explicit value  # noqa: PLR2004

    def test_transform_execution(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test that transforms resolve dependencies and execute correctly."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        def fake_import_from_string(object_path: str, default_module: str | None = None):
            assert object_path == "double_values"
            assert default_module == "gwmock_pop.transforms"

            def fake_transform(*, values: jnp.ndarray, scale: int) -> jnp.ndarray:
                return values * scale

            return fake_transform

        simulator = GraphSimulator(config=config)
        simulator._sampled_values["mass_1"] = jnp.array([1.0, 2.0, 3.0])
        monkeypatch.setattr(graph_module, "import_from_string", fake_import_from_string)

        result = simulator._execute_transform(
            {
                "function": "double_values",
                "arguments": {"values": "@mass_1", "scale": 2},
            }
        )

        assert jnp.array_equal(result, jnp.array([2.0, 4.0, 6.0]))

    def test_transform_execution_missing_dependency(self) -> None:
        """Test that transforms raise when a dependency has not been sampled."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)

        with pytest.raises(ValueError, match="Dependency 'mass_1' not available for transform"):
            simulator._execute_transform(
                {
                    "function": "double_values",
                    "arguments": {"values": "@mass_1"},
                }
            )

    def test_simulate_with_transform(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test simulation with both sampler and transform."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                        "n_samples": 20,
                    },
                },
            },
            "mass_1_transformed": {
                "transform": {
                    "function": "offset_values",
                    "arguments": {"values": "@mass_1"},
                },
            },
            "mass_2": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                        "n_samples": 20,
                    },
                },
            },
        }

        real_import_from_string = graph_module.import_from_string

        def fake_import_from_string(object_path: str, default_module: str | None = None):
            if object_path == "offset_values":
                assert default_module == "gwmock_pop.transforms"

                def fake_transform(*, values: jnp.ndarray) -> jnp.ndarray:
                    return values + 1

                return fake_transform

            return real_import_from_string(object_path=object_path, default_module=default_module)

        simulator = GraphSimulator(config=config)
        monkeypatch.setattr(graph_module, "import_from_string", fake_import_from_string)
        monkeypatch.setattr(
            simulator,
            "_get_ordered_parameters",
            lambda: ["mass_1", "mass_1_transformed", "mass_2"],
        )
        result = simulator()

        assert result.shape == (20, 3)  # mass_1, mass_1_transformed, mass_2
        assert jnp.allclose(result[:, 1], result[:, 0] + 1)

    def test_reset_without_rng_manager(self) -> None:
        """Test that reset works even without _rng_manager attribute."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        _ = simulator()

        # Manually remove _rng_manager to test reset without it
        if hasattr(simulator, "_rng_manager"):
            delattr(simulator, "_rng_manager")

        # Reset should not raise error
        simulator.reset()
        assert len(simulator._sampled_values) == 0

    def test_execute_sampler_with_none_arguments(self) -> None:
        """Test that _execute_sampler handles None arguments correctly."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)

        # Test with None arguments - should be converted to empty dict
        # This tests the code path where arguments is None and gets converted to {}
        # We use a minimal sampler that doesn't require arguments
        def minimal_sampler(**kwargs: Any) -> jnp.ndarray:
            return jnp.array([1.0, 2.0, 3.0])

        # Monkeypatch import_from_string to return our minimal sampler
        monkeypatch = pytest.MonkeyPatch()
        monkeypatch.setattr(graph_module, "import_from_string", lambda *args, **kwargs: minimal_sampler)

        result = simulator._execute_sampler(
            "minimal_sampler",
            None,  # type: ignore[arg-type]
        )
        assert result is not None
        assert isinstance(result, jnp.ndarray)
        monkeypatch.undo()

    def test_execute_sampler_with_invalid_arguments_type(self) -> None:
        """Test that _execute_sampler raises error for invalid arguments type."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        # Test with string arguments - should raise ValueError
        with pytest.raises(ValueError, match="Sampler arguments must be a mapping"):
            simulator._execute_sampler(
                "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                "invalid_arguments",  # type: ignore[arg-type]
            )

    def test_execute_sampler_with_unresolved_dependency(self) -> None:
        """Test that _execute_sampler raises error for unresolved dependency."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        # Test with dependency that hasn't been sampled
        with pytest.raises(ValueError, match="Dependency 'nonexistent' not sampled yet"):
            simulator._execute_sampler(
                "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                {"alpha_1": "@nonexistent"},
            )

    def test_execute_transform_with_string_expression(self) -> None:
        """Test that _execute_transform raises error for string transform."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        # Test with string transform - should raise ValueError
        with pytest.raises(ValueError, match="String transform expressions are not supported"):
            simulator._execute_transform("@mass_1 * 2")  # type: ignore[arg-type]

    def test_execute_transform_with_none_arguments(self) -> None:
        """Test that _execute_transform handles None arguments correctly."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)

        # Mock transform that expects arguments
        def fake_transform(**kwargs: Any) -> jnp.ndarray:
            return jnp.array([1.0, 2.0, 3.0])

        # Monkeypatch import_from_string to return our fake transform
        monkeypatch = pytest.MonkeyPatch()
        monkeypatch.setattr(graph_module, "import_from_string", lambda *args, **kwargs: fake_transform)

        # Test with None arguments - should be converted to empty dict
        result = simulator._execute_transform(
            {
                "function": "fake_transform",
                "arguments": None,
            }
        )
        assert result is not None
        assert isinstance(result, jnp.ndarray)
        monkeypatch.undo()

    def test_execute_transform_with_invalid_arguments_type(self) -> None:
        """Test that _execute_transform raises error for invalid arguments type."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        # Test with string arguments - should raise ValueError
        with pytest.raises(ValueError, match="Only mapping-style transform arguments are currently supported"):
            simulator._execute_transform(
                {
                    "function": "some_transform",
                    "arguments": "invalid_arguments",
                }
            )

    def test_from_config_file_with_invalid_root(self) -> None:
        """Test that config file with non-dict root raises ValueError."""
        # Create a YAML file with a list at root
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write("- item1\n- item2\n")
            temp_path = f.name

        try:
            with pytest.raises(ValueError, match="Config file must contain a mapping"):
                GraphSimulator.from_config_file(temp_path)
        finally:
            Path(temp_path).unlink()

    def test_from_config_file_with_list_root(self) -> None:
        """Test that config file with list root raises ValueError."""
        # Create a YAML file with a list at root (TOML cannot have list at root)
        with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml", delete=False) as f:
            f.write("- 1\n- 2\n- 3\n")
            temp_path = f.name

        try:
            with pytest.raises(ValueError, match="Config file must contain a mapping"):
                GraphSimulator.from_config_file(temp_path)
        finally:
            Path(temp_path).unlink()

    def test_coerce_output_column_scalar_raises_error(self) -> None:
        """Test that _coerce_output_column raises error for scalar output."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        # Test with scalar value - should raise ValueError
        with pytest.raises(ValueError, match="produced a scalar output"):
            simulator._coerce_output_column("mass_1", jnp.array(5.0), None)

    def test_coerce_output_column_sample_count_mismatch(self) -> None:
        """Test that _coerce_output_column raises error for sample count mismatch."""
        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
        }

        simulator = GraphSimulator(config=config)
        # Test with wrong sample count - should raise ValueError
        wrong_count_array = jnp.array([1.0, 2.0, 3.0])  # 3 samples instead of 100
        with pytest.raises(ValueError, match="produced 3 samples, expected 100"):
            simulator._coerce_output_column("mass_1", wrong_count_array, 100)

    def test_simulate_with_transform_branch(self, monkeypatch: pytest.MonkeyPatch) -> None:
        """Test that transform branch is executed correctly in _simulate_impl."""

        # Create a simple transform that accepts any kwargs
        def dummy_transform(**kwargs: Any) -> jnp.ndarray:
            return jnp.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0])

        # Monkeypatch to use our dummy transform
        monkeypatch.setattr(
            graph_module,
            "import_from_string",
            lambda object_path, default_module=None: dummy_transform,
        )

        config = {
            "mass_1": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 10,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
            },
            "mass_1_transformed": {
                "transform": {
                    "function": "dummy_transform",
                    "arguments": {"values": "@mass_1"},
                },
            },
        }

        simulator = GraphSimulator(config=config)
        # Just verify that the transform branch is reachable
        # The actual output is not important, we just want to ensure
        # the transform branch in _simulate_impl is executed
        result = simulator()

        # Should have 2 parameters: mass_1 and mass_1_transformed
        expected_n_parameters = 2
        assert result.shape[1] == expected_n_parameters
        assert "mass_1" in simulator.parameter_names
        assert "mass_1_transformed" in simulator.parameter_names

    def test_simulate_empty_parameter_names_raises_error(self) -> None:
        """Test that _simulate_impl raises error when no output parameters defined."""
        config = {
            "intermediate_param": {
                "sampler": {
                    "function": "gwmock_pop.samplers.planck_tapered_broken_power_law_plus_two_peaks",
                    "arguments": {
                        "n_samples": 100,
                        "alpha_1": 1.72,
                        "alpha_2": 4.51,
                        "transition": 35.6,
                        "minimum": 5.06,
                        "maximum": 300.0,
                        "mean_1": 9.76,
                        "sigma_1": 0.649,
                        "mean_2": 32.8,
                        "sigma_2": 3.92,
                        "taper_range": 4.32,
                        "lambda_0": 0.361,
                        "lambda_1": 0.586,
                    },
                },
                "exclude": True,  # Exclude from output
                "intermediate": True,  # Mark as intermediate
            },
        }

        simulator = GraphSimulator(config=config)
        # parameter_names will be empty because all params are excluded/intermediate
        assert simulator.parameter_names == []

        # Simulate should raise ValueError when no output parameters
        with pytest.raises(ValueError, match="does not define any output parameters"):
            simulator()

    def test_from_config_file_with_unsupported_suffix_other_error(self) -> None:
        """Test that from_config_file re-raises non-matching errors."""
        # Create a file with unsupported suffix that triggers a different error
        with tempfile.NamedTemporaryFile(mode="w", suffix=".txt", delete=False) as f:
            # Write content that will cause a different ValueError
            f.write("invalid yaml content {{")
            temp_path = f.name

        try:
            # This should raise a ValueError with unsupported suffix message
            with pytest.raises(ValueError, match="Suffix of config_path"):
                GraphSimulator.from_config_file(temp_path)
        finally:
            Path(temp_path).unlink()



"""Helper functions to extract the dependencies of a sampler."""

from __future__ import annotations

from typing import Any

from gwmock_pop.graph.generic import extract_dependencies_from_spec


def extract_sampler_dependencies(sampler_spec: dict[str, Any]) -> set[str]:
    """Extract the dependencies of the sampler.

    Args:
        sampler_spec: Specifications of the sampler.

    Returns:
        A set of dependencies.
    """
    dependencies = set()

    # Sampler is a dictionary.
    # Look for @references in its values
    if isinstance(sampler_spec, dict):
        # Special case: function-based sampler
        if "function" in sampler_spec:
            # Prefer "arguments" over "depends_on"
            args = sampler_spec.get("arguments")
            if args is None:
                # Fall back to depends_on if arguments is not present
                args = sampler_spec.get("depends_on", [])
            elif isinstance(args, list) and len(args) == 0:
                # Empty list means no arguments, fall back to depends_on
                args = sampler_spec.get("depends_on", [])
            elif isinstance(args, dict) and len(args) == 0:
                # Empty dict means no arguments, fall back to depends_on
                args = sampler_spec.get("depends_on", {})
            # Otherwise, use arguments as-is

            if isinstance(args, list):
                dependencies.update(
                    arg[1:] for arg in args if isinstance(arg, str) and arg.startswith("@") and len(arg) > 1
                )
            elif isinstance(args, dict):
                dependencies.update(
                    v[1:] for v in args.values() if isinstance(v, str) and v.startswith("@") and len(v) > 1
                )
        else:
            dependencies = extract_dependencies_from_spec(sampler_spec)
    return dependencies



"""Unit tests for BBHParameters dataclass."""

from types import MappingProxyType

import jax.numpy as jnp

from gwmock_pop.parameters.bbh import BBHParameters


class TestBBHParameters:
    """Test class for BBHParameters dataclass."""

    def test_bbh_parameters_initialization(self):
        """Test BBHParameters dataclass initialization."""
        # Test with scalar values
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.0),
            f_ref=jnp.array(20.0),
        )

        assert params.mass_1 == jnp.array(2.0)
        assert params.mass_2 == jnp.array(1.0)
        assert params.redshift == jnp.array(0.0)
        assert params.extra == {}

        # Test with array values
        params = BBHParameters(
            mass_1=jnp.array([2.0, 3.0]),
            mass_2=jnp.array([1.0, 2.0]),
            spin_1x=jnp.array([0.0, 0.0]),
            spin_1y=jnp.array([0.0, 0.0]),
            spin_1z=jnp.array([0.0, 0.0]),
            spin_2x=jnp.array([0.0, 0.0]),
            spin_2y=jnp.array([0.0, 0.0]),
            spin_2z=jnp.array([0.0, 0.0]),
            eccentricity=jnp.array([0.0, 0.0]),
            distance=jnp.array([100.0, 150.0]),
            coa_phase=jnp.array([0.0, 0.0]),
            inclination=jnp.array([0.0, 0.0]),
            theta_jn=jnp.array([0.0, 0.0]),
            long_asc_nodes=jnp.array([0.0, 0.0]),
            mean_per_ano=jnp.array([0.0, 0.0]),
            coa_time=jnp.array([0.0, 0.0]),
            right_ascension=jnp.array([0.0, 0.0]),
            declination=jnp.array([0.0, 0.0]),
            polarization_angle=jnp.array([0.0, 0.0]),
            redshift=jnp.array([0.0, 0.1]),
            f_ref=jnp.array([20.0, 20.0]),
        )

        assert jnp.array_equal(params.mass_1, jnp.array([2.0, 3.0]))
        assert jnp.array_equal(params.mass_2, jnp.array([1.0, 2.0]))
        assert jnp.array_equal(params.redshift, jnp.array([0.0, 0.1]))

    def test_bbh_parameters_chirp_mass(self):
        """Test chirp_mass property."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.0),
            f_ref=jnp.array(20.0),
        )

        result = params.chirp_mass
        expected = (2.0 * 1.0) ** 0.6 / (2.0 + 1.0) ** 0.2
        assert jnp.isclose(result, expected)

        # Test with arrays
        params = BBHParameters(
            mass_1=jnp.array([2.0, 3.0]),
            mass_2=jnp.array([1.0, 2.0]),
            spin_1x=jnp.array([0.0, 0.0]),
            spin_1y=jnp.array([0.0, 0.0]),
            spin_1z=jnp.array([0.0, 0.0]),
            spin_2x=jnp.array([0.0, 0.0]),
            spin_2y=jnp.array([0.0, 0.0]),
            spin_2z=jnp.array([0.0, 0.0]),
            eccentricity=jnp.array([0.0, 0.0]),
            distance=jnp.array([100.0, 100.0]),
            coa_phase=jnp.array([0.0, 0.0]),
            inclination=jnp.array([0.0, 0.0]),
            theta_jn=jnp.array([0.0, 0.0]),
            long_asc_nodes=jnp.array([0.0, 0.0]),
            mean_per_ano=jnp.array([0.0, 0.0]),
            coa_time=jnp.array([0.0, 0.0]),
            right_ascension=jnp.array([0.0, 0.0]),
            declination=jnp.array([0.0, 0.0]),
            polarization_angle=jnp.array([0.0, 0.0]),
            redshift=jnp.array([0.0, 0.0]),
            f_ref=jnp.array([20.0, 20.0]),
        )

        result = params.chirp_mass
        expected = jnp.array([(2.0 * 1.0) ** 0.6 / (2.0 + 1.0) ** 0.2, (3.0 * 2.0) ** 0.6 / (3.0 + 2.0) ** 0.2])
        assert jnp.allclose(result, expected)

    def test_bbh_parameters_symmetric_mass_ratio(self):
        """Test symmetric_mass_ratio property."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.0),
            f_ref=jnp.array(20.0),
        )

        result = params.symmetric_mass_ratio
        expected = (2.0 * 1.0) / (2.0 + 1.0) ** 2
        assert jnp.isclose(result, expected)

        # Test with equal masses (should be 0.25, the maximum)
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(2.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.0),
            f_ref=jnp.array(20.0),
        )

        result = params.symmetric_mass_ratio
        assert jnp.isclose(result, 0.25)

    def test_bbh_parameters_total_mass(self):
        """Test total_mass property."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.0),
            f_ref=jnp.array(20.0),
        )

        result = params.total_mass
        expected = 3.0
        assert jnp.isclose(result, expected)

    def test_bbh_parameters_mass_ratio(self):
        """Test mass_ratio property."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.0),
            f_ref=jnp.array(20.0),
        )

        result = params.mass_ratio
        expected = 0.5
        assert jnp.isclose(result, expected)

    def test_bbh_parameters_source_frame_mass_1(self):
        """Test source_frame_mass_1 property."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.5),
            f_ref=jnp.array(20.0),
        )

        result = params.source_frame_mass_1
        expected = 2.0 / (1.0 + 0.5)
        assert jnp.isclose(result, expected)

    def test_bbh_parameters_source_frame_mass_2(self):
        """Test source_frame_mass_2 property."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.5),
            f_ref=jnp.array(20.0),
        )

        result = params.source_frame_mass_2
        expected = 1.0 / (1.0 + 0.5)
        assert jnp.isclose(result, expected)

    def test_bbh_parameters_source_frame_chirp_mass(self):
        """Test source_frame_chirp_mass property."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.5),
            f_ref=jnp.array(20.0),
        )

        # First compute the chirp mass in detector frame
        detector_chirp_mass = (2.0 * 1.0) ** 0.6 / (2.0 + 1.0) ** 0.2
        result = params.source_frame_chirp_mass
        expected = detector_chirp_mass / (1.0 + 0.5)
        assert jnp.isclose(result, expected)

    def test_bbh_parameters_source_frame_total_mass(self):
        """Test source_frame_total_mass property."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.5),
            f_ref=jnp.array(20.0),
        )

        # First compute the total mass in detector frame
        detector_total_mass = 3.0
        result = params.source_frame_total_mass
        expected = detector_total_mass / (1.0 + 0.5)
        assert jnp.isclose(result, expected)

    def test_bbh_parameters_extra_field(self):
        """Test extra field functionality."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.0),
            f_ref=jnp.array(20.0),
            extra={"custom_param": "value"},
        )

        assert params.extra["custom_param"] == "value"

        # Test that extra field is immutable (uses MappingProxyType)
        # This should not raise an error during access, but should prevent modification
        assert hasattr(params.extra, "__getitem__")

    def test_bbh_parameters_extra_field_immutability(self):
        """Test that extra field is immutable."""
        params = BBHParameters(
            mass_1=jnp.array(2.0),
            mass_2=jnp.array(1.0),
            spin_1x=jnp.array(0.0),
            spin_1y=jnp.array(0.0),
            spin_1z=jnp.array(0.0),
            spin_2x=jnp.array(0.0),
            spin_2y=jnp.array(0.0),
            spin_2z=jnp.array(0.0),
            eccentricity=jnp.array(0.0),
            distance=jnp.array(100.0),
            coa_phase=jnp.array(0.0),
            inclination=jnp.array(0.0),
            theta_jn=jnp.array(0.0),
            long_asc_nodes=jnp.array(0.0),
            mean_per_ano=jnp.array(0.0),
            coa_time=jnp.array(0.0),
            right_ascension=jnp.array(0.0),
            declination=jnp.array(0.0),
            polarization_angle=jnp.array(0.0),
            redshift=jnp.array(0.0),
            f_ref=jnp.array(20.0),
            extra={"custom_param": "value"},
        )

        # Verify that we can access the extra field
        assert params.extra["custom_param"] == "value"

        # Test that the extra field is a MappingProxyType (immutable)

        assert isinstance(params.extra, MappingProxyType)
